""" **Description**

        An Infinite Impulse Response ( IIR ) filter realizes a discrete
        difference equation as a function of a recursive coefficient array,
        a forward coefficient array, and a state array of a specified order,
        consuming an incident signal and producing a reference signal.

        .. math::

            y_{n} = \sum_{i = 1}^{N} a_{i,n} y_{n-i} + \sum_{i = 0}^{N} b_{i,n} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i,n} b_{0,n} + b_{i,n}\ ) s_{i,n} + b_{0,n} x_{n}\qquad a_{0,n} = 0

        .. math::

            s_{1,n+1} = \sum_{i = 1}^{N} a_{i,n} s_{i,n} + x_{n}\qquad\qquad s_{i,n+1} = s_{i-1,n}

        A reset may minimize edge effects at a discontinuity by assuming
        persistent operation at a specified incident signal condition.

        .. math::

            s_{i,n} = \\frac{1.0 - b_{0,n}}{\sum_{i=1}^{N} a_{i,n} b_{0,n} + b_{i,n}}\ x_{n}

        A frequency response is expressed as a function of a recursive
        coefficient array and a forward coefficient array.

        .. math::

            H_{z,n} = \\frac{\sum_{i = 0}^{N} b_{i,n} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i,n} z^{-i}}}

        A factory is defined to facilitate construction of an instance,
        defining a recursive coefficient array, a forward coefficient array,
        and a state array of a specified order, to realize specified
        constraints.  An instance, classification, frequency, order, count,
        complement, and gain are specified.

        Frequency corresponds to a -3 dB frequency response normalized relative
        to Nyquist.

        Classification is in ( 'Bessel', 'Butterworth', 'Chebyshev' ).

        * | 'Bessel' filters demonstrate maximally linear phase response or
          | constant group delay.

        * | 'Butterworth' filters demonstrate maximally flat magnitude response.

        * | 'Chebyshev' filters demonstrate minimally low magnitude response error
          | and improved rate of attenuation in a Type I form, with a maximum in
          | band ripple of 0.125 dB.

        Count is a quantity of filters of a specified order concatenated to
        form an aggregate frequency response in cascade form.

        Complement effectively constructs a mirror image of a specified
        frequency response.

    **Example**

        ::

            from diamondback import IirFilter
            import numpy

            # Create an instance from a Factory with constraints.

            obj = IirFilter.Factory.instance( typ = IirFilter, classification = 'Chebyshev', frequency = 0.1, order = 8, count = 1 )

            # Create an instance with coefficients.

            obj = IirFilter( a = obj.a, b = obj.b )

            # Estimate frequency response, group delay, and roots.

            y, f = obj.response( length = 8192, count = 1 )

            y, f = obj.delay( length = 8192, count = 1 )

            p, z = obj.roots( count = 1 )

            # Filter an incident signal.

            x = numpy.random.rand( 128 ) * 2.0 - 1.0

            obj.reset( x[ 0 ] )

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-23.

    **Definition**

"""

from diamondback.filters.FirFilter import FirFilter
from diamondback.interfaces.IA import IA
from diamondback.transforms.ZTransform import ZTransform
from typing import Any, List, Tuple, Union
import math
import numpy
import scipy.signal
import warnings

class IirFilter( FirFilter, IA ) :

    """ Infinite Impulse Response ( IIR ) filter.
    """

    class Factory( object ) :

        """ Factory.
        """

        _classification = ( 'Bessel', 'Butterworth', 'Chebyshev' )

        @staticmethod
        def _evaluate( classification : str, frequency : float, order : int ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

            """ Evaluates coefficients.

                Arguments :

                    classification : str - in ( 'Bessel', 'Butterworth', 'Chebyshev' ).

                    frequency : float - relative to Nyquist in ( 0.0, 1.0 ).

                    order : int.

                Returns :

                    a : numpy.ndarray - recursive coefficient.

                    b : numpy.ndarray - forward coefficient.
            """

            bilinear = True

            if ( classification == 'Bessel' ) :

                bilinear = False

                u, a = numpy.ones( 1 ), numpy.ones( 2 )

                for ii in range( 2, order + 1 ) :

                    x = numpy.concatenate( ( u, numpy.zeros( 2 ) ) ) + numpy.concatenate( ( [ 0.0 ], ( ( 2.0 * ii ) - 1.0 ) * a ) )

                    u, a = a, x

            elif ( classification == 'Butterworth' ) :

                a = numpy.ones( 1 )

                for ii in range( 1, ( order // 2 ) + 1 ) :

                    a = numpy.convolve( a, numpy.array( [ 1.0, -2.0 * math.cos( ( ( ( 2.0 * ii ) + order - 1.0 ) / ( 2.0 * order ) ) * math.pi ), 1.0 ] ) )

                if ( order & 1 ) :

                    a = numpy.convolve( a, numpy.ones( 2 ) )

            elif ( classification == 'Chebyshev' ) :

                ripple = 0.125

                u = numpy.array( [ numpy.exp( 1j * math.pi * x / ( 2.0 * order ) ) for x in range( 1, 2 * order, 2 ) ] )

                v = math.asinh( 1.0 / ( ( 10.0 ** ( 0.1 * ripple ) - 1.0 ) ** 0.5 ) ) / order

                a = ( numpy.poly( ( -math.sinh( v ) * u.imag + 1j * math.cosh( v ) * u.real ) * 2.0 * math.pi ) ).real

            a /= a[ -1 ]

            a, b = ZTransform.transform( a, [ 1.0 ], frequency, bilinear )

            b = numpy.poly( -numpy.ones( order ) )

            b *= ( 1.0 - sum( a ) ) / sum( b )

            return a, b

        @classmethod
        def instance( cls, typ : type, classification : str, frequency : float, order : int, count : int = 1, complement : bool = False, gain : float = 1.0 ) -> Any :

            """ Constructs an instance.

                Arguments :

                    typ : type - derived from IirFilter.

                    classification : str - in ( 'Bessel', 'Butterworth', 'Chebyshev' ).

                    frequency : float - relative to Nyquist in ( 0.0, 1.0 ).

                    order : int.

                    count : int.

                    complement : bool.

                    gain : float.

                Returns :

                    instance : typ( ).
            """

            if ( ( not typ ) or ( not issubclass( typ, IirFilter ) ) ) :

                raise ValueError( f'Type = {typ}' )

            if ( ( not classification ) or ( classification not in IirFilter.Factory._classification ) ) :

                raise ValueError( f'Classification = {classification}' )

            if ( ( frequency <= 0.0 ) or ( frequency >= 1.0 ) ) :

                raise ValueError( f'Frequency = {frequency}' )

            if ( order <= 0 ) :

                raise ValueError( f'Order = {order}' )

            if ( count <= 0 ) :

                raise ValueError( f'Count = {count}' )

            if ( complement ) :

                frequency = 1.0 - frequency

            beta, eps, error = 10.0, numpy.finfo( float ).eps, float( 'inf' )

            index, mu, zeta = 500 * ( 1 + ( count > 2 ) ), 2.5e-2, 1.0

            a, b = [ ], [ ]

            for ii in range( 0, index ) :

                u, v = IirFilter.Factory._evaluate( classification, zeta * frequency, order )

                x = numpy.exp( 1j * math.pi * frequency )

                e = ( 2.0 ** ( -0.5 ) ) - ( ( abs( numpy.polyval( v, x ) / numpy.polyval( numpy.concatenate( ( [ 1.0 ], -u[ 1 : ] ) ), x ) ) ) ** count )

                if ( abs( e ) < error ) :

                    a, b, error = u, v, abs( e )

                    if ( error < ( 10.0 * eps ) ) :

                        break

                zeta = max( zeta + mu * math.tanh( beta * e ), eps )

            if ( complement ) :

                a *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( a ) ) ] )

                b *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] )

                b /= sum( b * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] ) ) / sum( numpy.concatenate( ( [ 1.0 ], -a[ 1 : ] ) ) * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( a ) ) ] ) )

            return typ( a, b * gain )

    def __init__( self, a : Union[ List, numpy.ndarray ] = numpy.zeros( 1 ), b : Union[ List, numpy.ndarray ] = numpy.ones( 1 ), s : Union[ List, numpy.ndarray ] = numpy.zeros( 1 ) ) -> None :

        """ Initialize.

            Arguments :

                a : Union[ List, numpy.ndarray ] - recursive coefficient, s-domain.

                b : Union[ List, numpy.ndarray ] - forward coefficient.

                s : Union[ List, numpy.ndarray ] - state.
        """

        if ( ( not numpy.isscalar( a ) ) and ( not isinstance( a, numpy.ndarray ) ) ) :

            a = numpy.array( list( a ) )

        if ( ( len( a.shape ) != 1 ) or ( ( len( a ) > 0 ) and ( a[ 0 ] ) ) ) :

            raise ValueError( f'A = {a}' )

        if ( len( a ) < len( b ) ) :

            a = numpy.concatenate( ( a, numpy.zeros( len( b ) - len( a ) ) ) )

        if ( len( b ) < len( a ) ) :

            b = numpy.concatenate( ( b, numpy.zeros( len( a ) - len( b ) ) ) )

        if ( ( isinstance( a[ 0 ], complex ) ) or ( isinstance( b[ 0 ], complex ) ) ) :

            a, b = numpy.array( a, complex ), numpy.array( b, complex )

        if ( a[ 0 ] != 0.0 ) :

            raise ValueError( f'A = {a}' )

        super( ).__init__( b, s )

        self.a = numpy.array( a )

    def delay( self, length : int = 8192, count : int = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates group delay and produces a reference signal.

            Arguments :

                length : int.

                count : int.

            Returns :

                y : numpy.ndarray - reference signal.

                f : numpy.ndarray - relative to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length <= 0 ) :

            raise ValueError( f'Length = {length}' )

        if ( count <= 0 ) :

            raise ValueError( f'Count = {count}' )

        with warnings.catch_warnings( ) :

            warnings.simplefilter( 'ignore' )

            y, f = scipy.signal.group_delay( ( self.b, numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ) ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )

            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) )

        if ( length > 2 ) :

            y[ 0 ] = y[ 1 ] * 2.0 - y[ 2 ]

        return y, f

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( f'X = {x}' )

        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )

        for ii in range( 0, len( x ) ) :

            self.s[ 0 ] = x[ ii ]

            c = self.a * self.b[ 0 ] + self.b

            y[ ii ] = c.dot( self.s )

            if ( len( self.s ) > 1 ) :

                z = self.a.dot( self.s )

                self.s[ 1 : ] = self.s[ : -1 ]

                self.s[ 1 ] += z

        return y

    def reset( self, x : Union[ complex, float ] ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified incident signal condition.

            Arguments :

                x : Union[ complex, float ] - incident signal.
        """

        if ( not numpy.isscalar( x ) ) :

            raise ValueError( f'X = {x}' )

        if ( len( self.s ) > 1 ) :

            self.s.fill( x * ( 1.0 - self.b[ 0 ] ) / ( self.a[ 1 : ] * self.b[ 0 ] + self.b[ 1 : ] ).sum( ) )

    def response( self, length = 8192, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates frequency response and produces a reference signal.

            Arguments :

                length : int.

                count : int.

            Returns :

                y : numpy.ndarray - reference signal.

                f : numpy.ndarray - relative to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length <= 0 ) :

            raise ValueError( f'Length = {length}' )

        if ( count <= 0 ) :

            raise ValueError( f'Count = {count}' )

        with warnings.catch_warnings( ) :

            warnings.simplefilter( 'ignore' )

            y, f = scipy.signal.freqz( self.b, numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )

            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) ** count

        return y, f

    def roots( self, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates roots of a frequency response in poles and zeros.

            Arguments :

                count : int.

            Returns :

                p : numpy.ndarray - poles.

                z : numpy.ndarray - zeros.
        """

        p, z = numpy.tile( numpy.roots( numpy.concatenate( ( [ 1.0 ], -self.a[ 1 : ] ) ) ), count ), numpy.tile( numpy.roots( self.b ), count )

        return p[ numpy.argsort( abs( p ) ) ], z[ numpy.argsort( abs( z ) ) ]



""" **Description**

        Initialize.

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-22.

    **Definition**

"""

from .DiversityModel import DiversityModel  # noqa: F401
from .PrincipalComponentModel import PrincipalComponentModel  # noqa: F401



""" **Description**

        A rank filter realizes a nonlinear morphological operator consuming an
        incident signal, sorting, indexing, and selecting over a sliding window
        of a specified order, and producing a reference signal.

        Simple morphological operations include dilation, which defines a rank
        index equal to zero, erosion, which defines a rank index equal to order,
        and median, which defines a rank index equal to order divided by two
        for even order.  Compound morphological operations include close,
        consisting of sequential dilation and erosion, and open, consisting of
        sequential erosion and dilation.  An order and rank are specified.

        .. math::

            y_{n} = sort(\ x_{n-N+1\ :\ n}\ )[\ i\ ]

    **Example**

        ::

            from diamondback import RankFilter
            import numpy

            # Create an instance with rank and order.

            obj = RankFilter( rank = 4, order = 4 )

            # Filter an incident signal.

            x = numpy.concatenate( ( numpy.ones( 1 ), numpy.zeros( 10 ), numpy.ones( 4 ), numpy.zeros( 2 ), numpy.ones( 5 ), numpy.zeros( 6 ) ) )

            obj.reset( x[ 0 ] )

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.filters.FirFilter import FirFilter
from typing import Any, List, Union
import numpy

class RankFilter( FirFilter ) :

    """ Rank filter.
    """

    @property
    def rank( self ) :

        """ rank : int - in [ 0, order ].
        """

        return self._rank

    @rank.setter
    def rank( self, rank : int ) :

        if ( ( rank < 0 ) or ( rank > ( len( self.s ) - 1 ) ) ) :

            raise ValueError( f'Rank = {rank}' )

        self._rank = rank

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self.rank == other.rank ) )

    def __init__( self, rank : int, order : int ) -> None :

        """ Initialize.

            Arguments :

                rank : int - in [ 0, order ].

                order : int.
        """

        if ( ( rank < 0 ) or ( rank > order ) ) :

            raise ValueError( f'Rank = {rank} Order = {order}' )

        super( ).__init__( numpy.ones( order + 1 ) / ( order + 1 ) )

        self._rank = rank

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( f'X = {x}' )

        y = numpy.zeros( len( x ), type( x[ 0 ] ) )

        for ii in range( 0, len( x ) ) :

            self.s[ 0 ] = x[ ii ]

            y[ ii ] = self.s[ numpy.argsort( abs( self.s ) ) ][ self.rank ]

            if ( len( self.s ) > 1 ) :

                self.s[ 1 : ] = self.s[ : -1 ]

        return y



""" **Description**

        Initialize.

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-22.

    **Definition**

"""

from .IA import IA  # noqa: F401
from .IB import IB  # noqa: F401
from .IClear import IClear  # noqa: F401
from .IConfigure import IConfigure  # noqa: F401
from .ICount import ICount  # noqa: F401
from .IData import IData  # noqa: F401
from .IDate import IDate  # noqa: F401
from .IDispose import IDispose  # noqa: F401
from .IDuration import IDuration  # noqa: F401
from .IEqual import IEqual  # noqa: F401
from .IFrequency import IFrequency  # noqa: F401
from .IIdentity import IIdentity  # noqa: F401
from .IInterval import IInterval  # noqa: F401
from .ILabel import ILabel  # noqa: F401
from .ILatency import ILatency  # noqa: F401
from .ILive import ILive  # noqa: F401
from .IModel import IModel  # noqa: F401
from .IPath import IPath  # noqa: F401
from .IPeriod import IPeriod  # noqa: F401
from .IPhase import IPhase  # noqa: F401
from .IProxy import IProxy  # noqa: F401
from .IQ import IQ  # noqa: F401
from .IRate import IRate  # noqa: F401
from .IReady import IReady  # noqa: F401
from .IReset import IReset  # noqa: F401
from .IResolution import IResolution  # noqa: F401
from .IRotation import IRotation  # noqa: F401
from .IS import IS  # noqa: F401
from .IStream import IStream  # noqa: F401
from .ITimeOut import ITimeOut  # noqa: F401
from .ITimeZone import ITimeZone  # noqa: F401
from .IUrl import IUrl  # noqa: F401
from .IValid import IValid  # noqa: F401
from .IVersion import IVersion  # noqa: F401



""" **Description**

        A Goertzel filter realizes a discrete difference equation which
        approximates a discrete Fourier transform evaluated at a specified
        normalized frequency and order, consuming an incident signal and
        producing a reference signal.

        .. math::

            y_{n} = \sum_{i = 1}^{N} a_{i} y_{n-i} + \sum_{i = 0}^{N} b_{i} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i} b_{0} + b_{i}\ ) s_{i,n} + b_{0} x_{n}\qquad a_{0} = 0

        .. math::

            s_{1,n+1} = \sum_{i = 1}^{N} a_{i} s_{i,n} + x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        .. math::

            \matrix{ a = \scriptsize{ [\ \matrix{ 0 & 2\ \cos(\ \pi\ f\ ) & -1 }\ ] } & b = \scriptsize{ [\ \matrix{ 1 & -e^{\ j\ \pi\ f\ } & 0 } }\ ] }

        At the terminus of each window length a reference signal is evaluated
        to estimate a discrete Fourier transform at a specified normalized
        frequency.

        .. math::

            H_{z} = \\frac{\sum_{i = 0}^{N} b_{i} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i} z^{-i}}}

        A Goertzel filter is normalized by incident signal length.  An incident
        signal length is is inversely proportional to a normalized frequency
        resolution.

        .. math::

            N = \\frac{2}{R}

    **Example**

        ::

            from diamondback import ComplexExponentialFilter, GoertzelFilter
            import numpy

            b = WindowFilter.Factory.instance( WindowFilter, 'Hann', 128 ).b

            frequency = 0.1

            # Create an instance from coefficients and frequency.

            obj = GoertzelFilter( b = b, frequency = frequency )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( 1024 ) * frequency ) * numpy.random.rand( 1 )[ 0 ]

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-16.

    **Definition**

"""

from diamondback.interfaces.IFrequency import IFrequency
from diamondback.filters.IirFilter import IirFilter
from typing import Any, List, Union
import math
import numpy

class GoertzelFilter( IirFilter, IFrequency ) :

    """ Goertzel filter.
    """

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self._index == other._index ) and ( numpy.allclose( self._w, other._w ) ) )

    def __init__( self, b : Union[ List, numpy.ndarray ], frequency : float ) -> None :

        """ Initialize.

            Arguments :

                b : Union[ List, numpy.ndarray ] - forward coefficient.

                frequency : float - relative to Nyquist in [ -1.0, 1.0 ).
        """

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :

            b = numpy.array( list( b ) )

        if ( ( len( b.shape ) != 1 ) or ( len( b ) == 0 ) or ( not b.any( ) ) ) :

            raise ValueError( f'B = {b}' )

        u = numpy.array( [ 0.0, 2.0 * math.cos( math.pi * frequency ), -1.0 ] )

        v = numpy.array( [ 1.0, -numpy.exp( -1j * math.pi * frequency ), 0.0 ] )

        super( ).__init__( u, v )

        self._index, self._w = 0, numpy.array( b )

        self.frequency = frequency

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( f'X = {x}' )

        y = numpy.zeros( int( numpy.ceil( len( x ) / len( self._w ) ) ) + 1, complex )

        u = ( 1.0 + int( not isinstance( x[ 0 ], complex ) ) ) / len( self._w )

        jj = 0

        for ii in range( 0, len( x ) ) :

            v = super( ).filter( u * self._w[ self._index ] * x[ ii : ii + 1 ] )

            self._index += 1

            if ( self._index >= len( self._w ) ) :

                y[ jj ] = v[ 0 ]

                self.s[ : ] = 0.0

                self._index = 0

                jj += 1

        if ( jj != len( y ) ) :

            y = y[ : jj ]

        return y



""" **Description**

        Initialize.

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-22.

    **Definition**

"""

from .commons import Log  # noqa: F401
from .commons import RestClient  # noqa: F401
from .commons import Serial  # noqa: F401
from .filters import FirFilter  # noqa: F401
from .filters import IirFilter  # noqa: F401
from .filters import ComplexBandPassFilter  # noqa: F401
from .filters import ComplexExponentialFilter  # noqa: F401
from .filters import ComplexFrequencyFilter  # noqa: F401
from .filters import DerivativeFilter  # noqa: F401
from .filters import GoertzelFilter  # noqa: F401
from .filters import IntegralFilter  # noqa: F401
from .filters import PidFilter  # noqa: F401
from .filters import PolynomialRateFilter  # noqa: F401
from .filters import PolyphaseRateFilter  # noqa: F401
from .filters import RankFilter  # noqa: F401
from .filters import WindowFilter  # noqa: F401
from .interfaces import IA  # noqa: F401
from .interfaces import IB  # noqa: F401
from .interfaces import IClear  # noqa: F401
from .interfaces import IConfigure  # noqa: F401
from .interfaces import ICount  # noqa: F401
from .interfaces import IData  # noqa: F401
from .interfaces import IDate  # noqa: F401
from .interfaces import IDispose  # noqa: F401
from .interfaces import IDuration  # noqa: F401
from .interfaces import IEqual  # noqa: F401
from .interfaces import IFrequency  # noqa: F401
from .interfaces import IIdentity  # noqa: F401
from .interfaces import IInterval  # noqa: F401
from .interfaces import ILabel  # noqa: F401
from .interfaces import ILatency  # noqa: F401
from .interfaces import ILive  # noqa: F401
from .interfaces import IModel  # noqa: F401
from .interfaces import IPath  # noqa: F401
from .interfaces import IPeriod  # noqa: F401
from .interfaces import IPhase  # noqa: F401
from .interfaces import IProxy  # noqa: F401
from .interfaces import IQ  # noqa: F401
from .interfaces import IRate  # noqa: F401
from .interfaces import IReady  # noqa: F401
from .interfaces import IReset  # noqa: F401
from .interfaces import IResolution  # noqa: F401
from .interfaces import IRotation  # noqa: F401
from .interfaces import IS  # noqa: F401
from .interfaces import IStream  # noqa: F401
from .interfaces import ITimeOut  # noqa: F401
from .interfaces import ITimeZone  # noqa: F401
from .interfaces import IUrl  # noqa: F401
from .interfaces import IValid  # noqa: F401
from .interfaces import IVersion  # noqa: F401
from .models import DiversityModel  # noqa: F401
from .models import PrincipalComponentModel  # noqa: F401
from .transforms import ComplexTransform  # noqa: F401
from .transforms import FourierTransform  # noqa: F401
from .transforms import PowerSpectrumTransform  # noqa: F401
from .transforms import WaveletTransform  # noqa: F401
from .transforms import ZTransform  # noqa: F401

__version__ = '1.0.99'



""" **Description**

        Reset interface.

    **Example**

        ::

            from diamondback import IReset, IS
            from typing import Union
            import numpy

            class Test( IReset, IS ) :

                def reset( self, x : Union[ complex, float ] ) -> None :

                    self.s[ : ] = x

            test = Test( )

            test.s = numpy.array( [ 0.0, 1.0 ] )

            test.reset( 0.5 )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-12.

    **Definition**

"""

from abc import ABC, abstractmethod
from typing import Union

class IReset( ABC ) :

    """ Reset interface.
    """

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

    @abstractmethod
    def reset( self, x : Union[ complex, float ] ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified incident signal condition.

            Arguments :

                x : Union[ complex, float ] - incident signal.
        """

        pass




""" **Description**

        Period interface.

    **Example**

        ::

            from diamondback import IPeriod

            class Test( IPeriod ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.period = 0.0

            test = Test( )

            test.period = 300.0

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any
import numpy

class IPeriod( IEqual ) :

    """ Period interface.
    """

    @property
    def period( self ) :

        """ period : float - in seconds in [ 0.0, inf ).
        """

        return self._period

    @period.setter
    def period( self, period : float ) :

        if ( period < 0.0 ) :

            raise ValueError( f'Period = {period}' )

        self._period = period

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.period, other.period ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._period = 0.0



""" **Description**

        Initialize.

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-22.

    **Definition**

"""

from .FirFilter import FirFilter  # noqa: F401
from .IirFilter import IirFilter  # noqa: F401
from .ComplexBandPassFilter import ComplexBandPassFilter  # noqa: F401
from .ComplexExponentialFilter import ComplexExponentialFilter  # noqa: F401
from .ComplexFrequencyFilter import ComplexFrequencyFilter  # noqa: F401
from .DerivativeFilter import DerivativeFilter  # noqa: F401
from .GoertzelFilter import GoertzelFilter  # noqa: F401
from .IntegralFilter import IntegralFilter  # noqa: F401
from .PidFilter import PidFilter  # noqa: F401
from .PolynomialRateFilter import PolynomialRateFilter  # noqa: F401
from .PolyphaseRateFilter import PolyphaseRateFilter  # noqa: F401
from .RankFilter import RankFilter  # noqa: F401
from .WindowFilter import WindowFilter  # noqa: F401




""" **Description**

        A polynomial rate filter produces a reference signal which approximates
        an incident signal evaluated at an effective frequency equal to the
        product of an incident sample frequency and a specified rate.

        A polynomial rate filter synthesizes a sequence of polynomials which
        form local approximations to an incident signal, and are evaluated at
        indices corresponding to a specified rate to produce a reference
        signal, effectively modifying the sampling rate by a specified rate
        ratio.

        A specified rate must be greater than zero, supporting decimation and
        interpolation.

        Latency compensation is not necessary, as no group delay is introduced.

        Edge effects are internally mitigated by linear extension of an
        incident signal.

        A polynomial rate filter may be the most appropriate option in
        applications which require fractional decimation and interpolation and
        benefit from minimization of edge effects due to discontinuous
        operation or dynamic rate.

    **Example**

        ::

            from diamondback import ComplexExponentialFilter, PolynomialRateFilter
            import math
            import numpy

            # Create an instance with rate and order.

            obj = PolynomialRateFilter( rate = math.pi, order = 3 )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( 128 ) * 0.1 ).real

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-03-19.

    **Definition**

"""

from diamondback.interfaces.IRate import IRate
from typing import Any, List, Union
import numpy

class PolynomialRateFilter( IRate ) :

    """ Polynomial rate filter.
    """

    @property
    def order( self ) :

        """ order : int - in [ 2, inf ).
        """

        return self._order

    @order.setter
    def order( self, order : int ) :

        if ( order < 2 ) :

            raise ValueError( f'Order = {order}' )

        self._order = order

    @IRate.rate.setter
    def rate( self, rate : float ) :

        """ rate : float - ratio of effective frequency in ( 0.0, inf ).
        """

        if ( rate <= 0.0 ) :

            raise ValueError( f'Rate = {rate}' )

        if ( not numpy.isclose( self.rate, rate ) ) :

            self._index = 0.0

        IRate.rate.fset( self, rate )

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self._index, other._index ) ) and ( self.order == other.order ) )

    def __init__( self, rate : float, order : int = 3 ) -> None :

        """ Initialize.

            Arguments :

                rate : float - ratio of effective frequency in [ 1.0, inf ).

                order : int - in [ 2 , inf ).
        """

        if ( order < 2 ) :

            raise ValueError( f'Order = {order}' )

        super( ).__init__( )

        self._index, self._order = 0.0, order

        self.rate = rate

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) < 2 ) ) :

            raise ValueError( f'X = {x}' )

        cc = len( x )

        x = numpy.concatenate( ( [ 2.0 * x[ 0 ] - x[ 1 ] ], x, [ 2.0 * x[ -1 ] - x[ -2 ], 3.0 * x[ -1 ] - 2.0 * x[ -2 ] ] ) )

        y = numpy.zeros( int( numpy.ceil( cc * self.rate ) ) )

        eps, u, v = numpy.finfo( float ).eps, numpy.linspace( -1.0, 2.0, 4 ), 1.0 / self.rate

        ii, jj = 0, 0

        while ( ii < cc ) :

            if ( self._index < ( 1.0 - eps ) ) :

                b = numpy.polyfit( u, x[ ii : ii + 4 ], self.order )

                while ( ( self._index < ( 1.0 - eps ) ) and ( jj < len( y ) ) ) :

                    y[ jj ] = numpy.polyval( b, self._index )

                    self._index += v

                    jj += 1

            self._index -= 1.0

            ii += 1

        return y[ 0 : min( jj, len( y ) ) ]



""" **Description**

        Equal interface.

    **Example**

        ::

            from diamondback import IEqual, IPhase
            from typing import Any

            class Test( IEqual, IPhase ) :

                def __eq__( self, other : Any ) -> bool :

                    return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.phase, other.phase ) ) )

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-23.

    **Definition**

"""

from typing import Any

class IEqual( object ) :

    """ Equal interface.
    """

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( isinstance( other, self.__class__ ) ) and ( ( id( self ) == id( other ) ) or ( super( ).__eq__( other ) ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        pass

    def __ne__( self, other : Any ) -> bool :

        """ Not equal.

            Arguments :

                other : Any.

            Returns :

                notequal : bool.
        """

        return not ( self == other )



""" **Description**

        Url interface.

    **Example**

        ::

            from diamondback import IUrl

            class Test( IUrl ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.url = 'http://127.0.0.1:8080/service'

            test = Test( )

            test.url = 'http://10.0.0.1:8080/service'

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-09-25.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any

class IUrl( IEqual ) :

    """ Url interface.
    """

    @property
    def url( self ) :

        """ url : str.
        """

        return self._url

    @url.setter
    def url( self, url : str ) :

        """ Url.
        """

        if ( url ) :

            url = url.strip( '/' )

        self._url = url

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self.url == other.url ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._url = 'http://127.0.0.1:8080'



""" **Description**

        State derivative interface.

    **Example**

        ::

            from diamondback import IQ
            import numpy

            class Test( IQ ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.q = numpy.array( [ 0.0, 0.1 ] )

            test = Test( )

            test.q[ : ] = 0.1

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any, List, Union
import numpy

class IQ( IEqual ) :

    """ State derivative interface.
    """

    @property
    def q( self ) :

        """ q : Union[ List, numpy.ndarray ] - state derivative.
        """

        return self._q

    @q.setter
    def q( self, q : Union[ List, numpy.ndarray ] ) :

        self._q = q

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.allclose( self.q, other.q ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._q = [ ]



"""" **Description**

        Date interface.

    **Example**

        ::

            from diamondback import IDate
            import datetime

            class Test( IDate ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.date = datetime.datetime.utcnow( ).replace( microsecond = 0, tzinfo = datetime.timezone.utc )

            test = Test( )

            test.date += datetime.timedelta( hours = 4 )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any
import datetime

class IDate( IEqual ) :

    """ Date interface.
    """

    @property
    def date( self ) :

        """ date : datetime.datetime.
        """

        return self._date

    @date.setter
    def date( self, date : datetime.datetime ) :

        if ( not date ) :

            raise ValueError( f'Date = {date}' )

        if ( not date.tzinfo ) :

            date = date.replace( tzinfo = datetime.timezone.utc )

        self._date = date.replace( microsecond = 0 )

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self.date == other.date ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._date = datetime.datetime.utcnow( ).replace( microsecond = 0, tzinfo = datetime.timezone.utc )



""" **Description**

    REST client instances define a client for simple REST service requests
    using the requests package.  An API and an elective dictionary of parameter
    strings are encoded to build a URL, elective binary or JSON data are
    defined in the body of a request, and a requests response containing JSON,
    text, or binary data is returned.

    Proxy, timeout, and URL definition are supported.

    Live makes a head request to a URL and detects a live service.

    **Example**

        ::

            from diamondback import RestClient
            from typing import Dict
            import numpy

            class TestClient( RestClient ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.proxy = { 'http' : '', 'https' : '' }

                def add( self, json : Dict[ str, numpy.ndarray ] ) -> numpy.ndarray :

                    return self.request( 'get', 'test/add', json = json ).json( )

            client = TestClient( )

            client.url = 'http://127.0.0.1:8080'

            client.timeout = ( 10.0, 60.0 )  # connect, read

            value = client.add( { 'x' : numpy.random.rand( 3 ), 'y' : numpy.random.rand( 3 ) } )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-10-22.

    **Definition**

"""

from diamondback.interfaces.ILive import ILive
from diamondback.interfaces.IProxy import IProxy
from diamondback.interfaces.ITimeOut import ITimeOut
from diamondback.interfaces.IUrl import IUrl
from typing import Any, Dict
import requests

class RestClient( ILive, IProxy, ITimeOut, IUrl ) :

    """ REST client.
    """

    @ILive.live.getter
    def live( self ) :

        """ live : bool.
        """

        try :

            requests.request( method = 'head', url = self.url, proxies = self.proxy, timeout = self.timeout )

            value = True

        except Exception :

            value = False

        return value

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self.proxy, self.timeout = { }, ( 10.0, 60.0 )

        self.url = 'http://127.0.0.1:8080'

    def request( self, method : str, api : str, item : Dict[ str, str ] = None, data : Any = None, json : Any = None ) -> requests.Response :

        """ Request client for simple REST service requests. An API and an
            elective dictionary of parameter strings are encoded to build a
            URL, elective binary or JSON data are defined in the body of a
            request, and a requests response containing JSON, text, or binary
            data is returned.

            Arguments :

                method : str - in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ).

                api : str - relative to the URL.

                item : Dict[ str, str ].

                data : Any.

                json : Any.

            Returns :

                value : requests.Response.
        """

        if ( not method ) :

            raise ValueError( f'Method = {method}' )

        method = method.lower( )

        if ( method not in ( 'delete', 'get', 'head', 'options', 'patch', 'post', 'put' ) ) :

            raise ValueError( f'Method = {method}' )

        if ( ( data ) and ( json ) ) :

            raise ValueError( f'Data = {data} JSON = {json}' )

        api = api.strip( '/' )

        url = self.url

        if ( api ) :

            url += '/' + api

        with requests.request( method = method, url = url, params = item, data = data, json = json, proxies = self.proxy, timeout = self.timeout ) as value :

            value.raise_for_status( )

        return value



""" **Description**

        Frequency interface.

    **Example**

        ::

            from diamondback import IFrequency

            class Test( IFrequency ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.frequency = 1.0

            test = Test( )

            test.frequency = 0.5

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any
import numpy

class IFrequency( IEqual ) :

    """ Frequency interface.
    """

    @property
    def frequency( self ) :

        """ frequency : float - relative to Nyquist in [ -1.0, 1.0 ].
        """

        return self._frequency

    @frequency.setter
    def frequency( self, frequency : float ) :

        if ( ( frequency < -1.0 ) or ( frequency > 1.0 ) ) :

            raise ValueError( f'Frequency = {frequency}' )

        self._frequency = frequency

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.frequency, other.frequency ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._frequency = 0.0



""" **Description**

        A Finite Impulse Response ( FIR ) filter realizes a discrete difference
        equation as a function of a forward coefficient array and a state array
        of a specified order, consuming an incident signal and producing a
        reference signal.

        .. math::

            y_{n} = \sum_{i = 0}^{N} b_{i,n} x_{n-i} = \sum_{i = 1}^{N} b_{i,n} s_{i,n} + b_{0,n} x_{n}

        .. math::

            s_{1,n+1} = x_{n}\qquad\quad s_{i,n+1} = s_{i-1,n}

        A reset may minimize edge effects at a discontinuity by assuming
        persistent operation at a specified incident signal condition.

        .. math::

            s_{i,n} = x_{n}

        A frequency response is expressed as a function of a forward
        coefficient array.

        .. math::

            H_{z,n} = \sum_{i = 0}^{N} b_{i,n} z^{-i}

        A factory is defined to facilitate construction of an instance,
        defining a forward coefficient array, and a state array of a
        specified order, to realize specified constraints.  An instance,
        classification, frequency, order, count, complement, and gain
        are specified.

        Frequency corresponds to a -3 dB frequency response normalized relative
        to Nyquist.

        Classification is in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

        * | 'Blackman' filters demonstrate low resolution and spectral leakage
          | with improved rate of attenuation.

        * | 'Hamming' filters demonstrate minimal nearest side lobe magnitude
          | response.

        * | 'Hann' filters demonstrate high resolution and spectral leakage.

        * | 'Kaiser' filters demonstrate flexible resolution and spectral
          | leakage dependent upon a beta value of a Bessel function of the
          | first kind, with beta equal to 7.0.

        Order must even to ensure a Type I form linear phase solution.

        Count is a quantity of filters of a specified order concatenated to
        form an aggregate frequency response in cascade form.

        Complement effectively constructs a mirror image of a specified
        frequency response.

    **Example**

        ::

            from diamondback import FirFilter
            import numpy

            # Create an instance from a Factory with constraints.

            obj = FirFilter.Factory.instance( typ = FirFilter, classification = 'Kaiser', frequency = 0.1, order = 32, count = 1 )

            # Create an instance with coefficients.

            obj = FirFilter( b = obj.b )

            # Estimate frequency response, group delay, and roots.

            y, f = obj.response( length = 8192, count = 1 )

            y, f = obj.delay( length = 8192, count = 1 )

            p, z = obj.roots( count = 1 )

            # Filter an incident signal.

            x = numpy.random.rand( 128 ) * 2.0 - 1.0

            obj.reset( x[ 0 ] )

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-23.

    **Definition**

"""

from diamondback.interfaces.IB import IB
from diamondback.interfaces.IReset import IReset
from diamondback.interfaces.IS import IS
from typing import Any, List, Tuple, Union
import math
import numpy
import scipy.signal
import warnings

class FirFilter( IB, IReset, IS ) :

    """ Finite Impulse Response ( FIR ) filter.
    """

    class Factory( object ) :

        """ Factory.
        """

        _classification = ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' )

        @classmethod
        def instance( cls, typ : type, classification : str, frequency : float, order : int, count : int = 1, complement : bool = False, gain : float = 1.0 ) -> Any :

            """ Constructs an instance.

                Arguments :

                    typ : type - derived from FirFilter.

                    classification : str - in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

                    frequency : float - relative to Nyquist in ( 0.0, 1.0 ).

                    order : int.

                    count : int.

                    complement : bool.

                    gain : float.

                Returns :

                    instance : typ( ).
            """

            if ( ( not typ ) or ( not issubclass( typ, FirFilter ) ) ) :

                raise ValueError( f'Type = {typ}' )

            if ( ( not classification ) or ( classification not in FirFilter.Factory._classification ) ) :

                raise ValueError( f'Classification = {classification}' )

            if ( ( frequency <= 0.0 ) or ( frequency >= 1.0 ) ) :

                raise ValueError( f'Frequency = {frequency}' )

            if ( order < 0 ) :

                raise ValueError( f'Order = {order}' )

            if ( count <= 0 ) :

                raise ValueError( f'Count = {count}' )

            if ( complement ) :

                frequency = 1.0 - frequency

            if ( classification == 'Kaiser' ) :

                window = ( classification.lower( ), 7.0 )

            else :

                window = classification.lower( )

            beta, eps, error = 10.0, numpy.finfo( float ).eps, float( 'inf' )

            index, mu, zeta = 500 * ( 1 + ( count > 2 ) ), 2.5e-2, 1.0

            for ii in range( 0, index ) :

                with warnings.catch_warnings( ) :

                    warnings.simplefilter( 'ignore' )

                    v = scipy.signal.firwin( order + 1, zeta * frequency, None, window, True, True, 1.0 )

                    if ( numpy.isnan( v ).any( ) ) :

                        raise ValueError( f'V = {v}' )

                    x = numpy.exp( 1j * math.pi * frequency )

                    e = ( 2.0 ** ( -0.5 ) ) - ( abs( numpy.polyval( v, x ) ) ** count )

                    if ( abs( e ) < error ) :

                        b, error = v, abs( e )

                        if ( error < ( 10.0 * eps ) ) :

                            break

                    zeta = numpy.maximum( zeta + mu * math.tanh( beta * e ), eps )

            if ( complement ) :

                b *= numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] )

                b /= sum( b * numpy.array( [ ( ( -1.0 ) ** x ) for x in range( 0, len( b ) ) ] ) )

            return typ( b * gain )

    def __init__( self, b : Union[ List, numpy.ndarray ] = numpy.ones( 1 ), s : Union[ List, numpy.ndarray ] = numpy.zeros( 1 ) ) -> None :

        """ Initialize.

            Arguments :

                b : Union[ List, numpy.ndarray ] - forward coefficient.

                s : Union[ List, numpy.ndarray ] - state.
        """

        if ( ( not numpy.isscalar( b ) ) and ( not isinstance( b, numpy.ndarray ) ) ) :

            b = numpy.array( list( b ) )

        if ( ( len( b.shape ) != 1 ) or ( len( b ) == 0 ) ) :

            raise ValueError( f'B = {b}' )

        if ( ( not numpy.isscalar( s ) ) and ( not isinstance( s, numpy.ndarray ) ) ) :

            s = numpy.array( list( s ) )

        if ( ( len( s.shape ) != 1 ) or ( len( s ) == 0 ) ) :

            raise ValueError( f'S = {s}' )

        if ( len( b ) < len( s ) ) :

            b = numpy.concatenate( ( b, numpy.zeros( len( s ) - len( b ) ) ) )

        if ( len( s ) < len( b ) ) :

            s = numpy.concatenate( ( s, numpy.zeros( len( b ) - len( s ) ) ) )

        super( ).__init__( )

        self.b, self.s = numpy.array( b ), numpy.array( s, type( b[ 0 ] ) )

    def delay( self, length : int = 8192, count : int = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates group delay and produces a reference signal.

            Arguments :

                length : int.

                count : int.

            Returns :

                y : numpy.ndarray - reference signal.

                f : numpy.ndarray - relative to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length <= 0 ) :

            raise ValueError( f'Length = {length}' )

        if ( count <= 0 ) :

            raise ValueError( f'Count = {count}' )

        with warnings.catch_warnings( ) :

            warnings.simplefilter( 'ignore' )

            y, f = scipy.signal.group_delay( ( self.b, [ 1.0 ] ), length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )

            y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) )

        if ( length > 2 ) :

            y[ 0 ] = y[ 1 ] * 2.0 - y[ 2 ]

        return y, f

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) ) :

            raise ValueError( f'X = {x}' )

        y = numpy.zeros( len( x ), type( self.b[ 0 ] ) )

        for ii in range( 0, len( x ) ) :

            self.s[ 0 ] = x[ ii ]

            y[ ii ] = self.b.dot( self.s )

            if ( len( self.s ) > 1 ) :

                self.s[ 1 : ] = self.s[ : -1 ]

        return y

    def reset( self, x : Union[ complex, float ] ) -> None :

        """ Modifies a state to minimize edge effects by assuming persistent
            operation at a specified incident signal condition.

            Arguments :

                x : Union[ complex, float ] - incident signal.
        """

        if ( not numpy.isscalar( x ) ) :

            raise ValueError( f'X = {x}' )

        self.s.fill( x )

    def response( self, length = 8192, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates frequency response and produces a reference signal.

            Arguments :

                length : int.

                count : int.

            Returns :

                y : numpy.ndarray - reference signal.

                f : numpy.ndarray - relative to Nyquist in [ -1.0, 1.0 ).
        """

        if ( length <= 0 ) :

            raise ValueError( f'Length = {length}' )

        if ( count <= 0 ) :

            raise ValueError( f'Count = {count}' )

        y, f = scipy.signal.freqz( self.b, [ 1.0, 0.0 ], length, True )[ 1 ], numpy.linspace( -1.0, 1.0 - 2.0 / length, length )

        y = numpy.concatenate( ( y[ len( y ) // 2 : ], y[ : len( y ) // 2 ] ) ) ** count

        return y, f

    def roots( self, count = 1 ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Estimates roots of a frequency response in poles and zeros.

            Arguments :

                count : int.

            Returns :

                p : numpy.ndarray - poles.

                z : numpy.ndarray - zeros.
        """

        z = numpy.tile( numpy.roots( self.b ), count )

        return numpy.zeros( count * ( len( self.b ) - 1 ) ), z[ numpy.argsort( abs( z ) ) ]



""" **Description**

        A window filter realizes a discrete difference equation as a function
        of a forward coefficient array of a specified order, consuming an
        incident signal and producing a reference signal.

        .. math::

            y_{n} = b_{n}\ x_{n}

        A factory is defined to facilitate construction of an instance,
        defining a forward coefficient array of a specified order.  An
        instance, classification, order, and normalization are specified.

        Classification is in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

        * | 'Blackman' filters demonstrate low resolution and spectral leakage
          | with improved rate of attenuation.

        * | 'Hamming' filters demonstrate minimal nearest side lobe magnitude
          | response.

        * | 'Hann' filters demonstrate high resolution and spectral leakage.

        * | 'Kaiser' filters demonstrate flexible resolution and spectral
          | leakage dependent upon a beta value of a Bessel function of the
          | first kind, with beta equal to 7.0.

        Normal condition scales a forward coefficient array to electively
        compensate for energy loss.

        .. math::

            b_{n} = b_{n}\ \\frac{ N }{ \sum_{0}^{N-1}\ |\ b_{n}\ |}

    **Example**

        ::

            from diamondback import ComplexExponentialFilter, WindowFilter
            import numpy

            # Create an instance from a Factory with constraints.

            obj = WindowFilter.Factory.instance( typ = WindowFilter, classification = 'Hann', order = 15, normal = True )

            # Filter an incident signal.

            x = ComplexExponentialFilter( 0.0 ).filter( numpy.ones( len( obj.b ) ) * 0.1 ).real

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-04-13.

    **Definition**

"""

from diamondback.interfaces.IB import IB
from diamondback.interfaces.IEqual import IEqual
from typing import Any, List, Union
import numpy
import scipy.signal

class WindowFilter( IB, IEqual ) :

    """ Window filter.
    """

    class Factory( object ) :

        """ Factory.
        """

        _classification = ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' )

        @classmethod
        def instance( cls, typ : type, classification : str, order : int, normal : bool = True ) -> Any :

            """ Constructs an instance.

                Arguments :

                    typ : type - derived from WindowFilter.

                    classification : str - in ( 'Blackman', 'Hamming', 'Hann', 'Kaiser' ).

                    order : int.

                    normal : bool.

                Returns :

                    instance : typ( ).
            """

            if ( ( not typ ) or ( not issubclass( typ, WindowFilter ) ) ) :

                raise ValueError( f'Type = {typ}' )

            if ( ( not classification ) or ( classification not in WindowFilter.Factory._classification ) ) :

                raise ValueError( f'Classification = {classification}' )

            if ( order < 0 ) :

                raise ValueError( f'Order = {order}' )

            if ( classification == 'Kaiser' ) :

                window = ( classification.lower( ), 7.0 )

            else :

                window = classification.lower( )

            b = scipy.signal.get_window( window, order + 1, False )

            if ( normal ) :

                b *= ( order + 1 ) / sum( abs( b ) )

            return typ( b )

    def __init__( self, b : Union[ List, numpy.ndarray ] = numpy.ones( 1 ) ) -> None :

        """ Initialize.

            Arguments :

                b : Union[ List, numpy.ndarray ] - forward coefficient.
        """

        if ( ( numpy.isscalar( b ) ) or ( len( b.shape ) != 1 ) or ( len( b ) == 0 ) or ( not b.any( ) ) ) :

            raise ValueError( f'B = {b}' )

        super( ).__init__( )

        self.b = numpy.array( b )

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) != len( self.b ) ) ) :

            raise ValueError( f'X = {x}' )

        return self.b * x



""" **Description**

        Dispose interface.

    **Example**

        ::

            from diamondback import IDispose

            class Test( IDispose ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.dispose = False

            test = Test( )

            test.dispose = True

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2020-10-22.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any

class IDispose( IEqual ) :

    """ Dispose interface.
    """

    @property
    def dispose( self ) :

        """ dispose : bool.
        """

        return self._dispose

    @dispose.setter
    def dispose( self, dispose : bool ) :

        self._dispose = dispose

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self.dispose == other.dispose ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._dispose = False



""" **Description**

        Latency interface.

    **Example**

        ::

            from diamondback import ILatency

            class Test( ILatency ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.latency = 0.0

            test = Test( )

            test.latency = 600.0

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-12.

    **Definition**
"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any
import numpy

class ILatency( IEqual ) :

    """ Latency interface.
    """

    @property
    def latency( self ) :

        """ latency : float - in seconds in [ 0.0, inf ).
        """

        return self._latency

    @latency.setter
    def latency( self, latency : float ) :

        if ( latency < 0.0 ) :

            raise ValueError( f'Latency = {latency}' )

        self._latency = latency

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.latency, other.latency ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._latency = 0.0



""" **Description**

        A complex band pass filter produces a complex exponential incident
        signal at a specified normalized frequency and adapts a forward complex
        coefficient to produce a reference signal, which estimates a component
        of interest in a primary signal.  A normalized frequency and rate of
        adaptation are specified.

        .. math::

            x_{n} = e^{\ j\ \pi\ \phi_{n}}

        .. math::

            \phi_{n+1} = \phi_{n} + f_{n}

        .. math::

            y_{n} = b_{n} x_{n}

        .. math::

            e_{n} = d_{n} - y_{n}

        .. math::

            b_{n+1} = b_{n} + \mu e_{n} x_{n}^{*}

    **Example**

        ::

            from diamondback import ComplexBandPassFilter, ComplexExponentialFilter
            import numpy

            frequency = 0.1

            x = numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + frequency

            # Create a primary signal.

            d = ComplexExponentialFilter( phase = numpy.random.rand( 1 )[ 0 ] * 2.0 - 1.0 ).filter( x )

            # Create an instance with frequency and rate.

            obj = ComplexBandPassFilter( frequency = frequency, rate = 5.0e-2 )

            # Filter a primary signal.

            obj.reset( d[ 0 ] )

            y, e, b = obj.filter( d )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.filters.ComplexExponentialFilter import ComplexExponentialFilter
from diamondback.filters.FirFilter import FirFilter
from diamondback.interfaces.IFrequency import IFrequency
from diamondback.interfaces.IRate import IRate
from typing import Any, List, Tuple, Union
import numpy

class ComplexBandPassFilter( FirFilter, IFrequency, IRate ) :

    """ Complex band pass filter.
    """

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self._complexexponentialfilter == other._complexexponentialfilter ) )

    def __init__( self, frequency : float, rate : float ) -> None :

        """ Initialize.

            Arguments :

                frequency : float - relative to Nyquist in [ -1.0, 1.0 ).

                rate : float - in [ 0.0, 1.0 ].
        """

        if ( ( rate < 0.0 ) or ( rate > 1.0 ) ) :

            raise ValueError( f'Rate = {rate}' )

        super( ).__init__( numpy.array( [ numpy.finfo( float ).eps + 0j ] ), numpy.zeros( 1, complex ) )

        self._complexexponentialfilter = ComplexExponentialFilter( )

        self.frequency, self.rate = frequency, rate

    def filter( self, d : Union[ List, numpy.ndarray ], x : Union[ List, numpy.ndarray ] = None ) -> Tuple[ numpy.ndarray, numpy.ndarray, numpy.ndarray ] :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                d : Union[ List, numpy.ndarray ] - primary signal.

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.

                e : numpy.ndarray - error signal.

                b : numpy.ndarray - forward coefficient.
        """

        if ( ( not numpy.isscalar( d ) ) and ( not isinstance( d, numpy.ndarray ) ) ) :

            d = numpy.array( list( d ) )

        if ( ( len( d.shape ) != 1 ) or ( len( d ) == 0 ) ) :

            raise ValueError( f'D = {d}' )

        x = self._complexexponentialfilter.filter( numpy.ones( len( d ) ) * self.frequency )

        y, e, b = numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex ), numpy.zeros( len( x ), complex )

        for ii in range( 0, len( x ) ) :

            y[ ii ] = x[ ii ] * self.b[ 0 ]

            e[ ii ] = d[ ii ] - y[ ii ]

            b[ ii ] = self.b[ 0 ]

            self.b[ 0 ] += self.rate * e[ ii ] * numpy.conjugate( x[ ii ] )

        if ( not isinstance( d[ 0 ], complex ) ) :

            y, e = y.real, e.real

        return y, e, b



""" **Description**

        A serial instance encodes and decodes an instance or collection in
        BSON or JSON, and generates SHA3-256 codes, using the jsonpickle
        package.

        BSON, Base-85 encoded gzip JSON, embeds a datetime context, and code
        will not be consistent or useful for validation.

        An instance may be an object or a collection, referenced by abstract or
        concrete types, and the instance will be correctly encoded and decoded,
        without custom encoding definitions.  BSON binary format is selected by
        electing to compress.  Encoding may be specified if an alternative to
        UTF-8 is required.

        Comments may be filtered from JSON by electing to clean.  Python style
        docstring and line comments are supported, though line comments must be
        terminated by a new line.

        Singleton.

        Thread safe.

    **Example**

        ::

            from diamondback import Serial
            import numpy
            import pandas

            # Encode and decode a dictionary instance in JSON.

            x = { 'a' : numpy.random.rand( count ), 'b' : list( numpy.random.rand( count ) ) }

            z = Serial.decode( Serial.encode( x ) )

            # Encode and decode a dictionary instance in BSON.

            y = Serial.encode( x, compress = True )

            z = Serial.decode( y, compress = True )

            # Encode and decode a pandas data frame in BSON.

            model = pandas.DataFrame( { 'fruit' : [ 'orange', 'apple', 'kiwi' ], 'Cost' : [ 1.25, 1.5, 0.30 ] } )

            y = Serial.encode( model )

            # Generate SHA3-256 code for an encoded model.

            code = Serial.code( y )

            z = Serial.decode( y )

            # Decode a dictionary instance from JSON.

            z = Serial.decode( '{ "a" : 1.0, "b" : 2.0, "c" : 3.14159 }' )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-07-13.

    **Definition**

"""

from typing import Any
import base64
import gzip
import hashlib
import jsonpickle
import jsonpickle.ext.numpy
import jsonpickle.ext.pandas
import re

class Serial( object ) :

    """ Serial.
    """

    jsonpickle.ext.numpy.register_handlers( )

    jsonpickle.ext.pandas.register_handlers( )

    @staticmethod
    def code( state : str, encoding : str = 'utf_8' ) -> str :

        """ Code generation.  SHA3-256 hash.

            Arguments :

                state : str.

                encoding : str.

            Returns :

                code : str.
        """

        if ( not state ) :

            raise ValueError( f'State = {state}' )

        if ( not encoding ) :

            raise ValueError( f'Encoding = {encoding}' )

        return hashlib.sha3_256( bytes( state, encoding ) ).hexdigest( )

    @staticmethod
    def decode( state : str, compress : bool = False, encoding : str = 'utf_8', clean : bool = False ) -> Any :

        """ Decodes an instance or collection from a BSON or JSON state.
            Encoding may be specified if an alternative to UTF-8 is required.
            Python style docstring and line comments may be cleaned, though
            line comments must be terminated by a new line.

            Arguments :

                state : str.

                compress : bool.

                encoding : str.

                clean : bool - clean comments.

            Returns :

                instance : Any.
        """

        if ( not state ) :

            raise ValueError( f'State = {state}' )

        if ( not encoding ) :

            raise ValueError( f'Encoding = {encoding}' )

        if ( compress ) :

            state = str( gzip.decompress( base64.b85decode( bytes( state, encoding ) ) ), encoding )

        if ( clean ) :

            state = re.sub( re.compile( '#.*?\n', re.DOTALL ), '', re.sub( re.compile( '\""".*?\"""', re.DOTALL ), '', state ) )

        return jsonpickle.decode( state )

    @staticmethod
    def encode( instance : Any, compress : bool = False, encoding : str = 'utf_8' ) -> str :

        """ Encodes BSON or JSON.  Encoding may be specified if an alternative
            to UTF-8 is required.

            Arguments :

                instance : Any.

                compress : bool.

                encoding : str.

            Returns :

                state : str.
        """

        if ( not encoding ) :

            raise ValueError( f'Encoding = {encoding}' )

        state = jsonpickle.encode( instance, separators = ( ',', ':' ) )

        if ( compress ) :

            state = str( base64.b85encode( gzip.compress( bytes( state, encoding ) ) ), encoding )

        return state



""" **Description**

        A z transform converts a continuous s-domain differential equation to a
        discrete z-domain difference equation as a function of a recursive
        coefficient array and a forward coefficient array of a specified order.
        A normalized frequency and bilinear condition are specified.

        Singleton.

        .. math::

            y_{n} = \sum_{i = 1}^{N} a_{i} y_{n-i} + \sum_{i = 0}^{N} b_{i} x_{n-i} = \sum_{i = 1}^{N} (\ a_{i} b_{0} + b_{i}\ ) s_{i,n} + b_{0} x_{n}\qquad a_{0} = 0

        A frequency response is expressed as a function of a recursive coefficient
        array and a forward coefficient array, in s-domain and z-domain.

        .. math::

            H_{s,n} = \\frac{\sum_{i = 0}^{N} v_{i} s^{N-i}}{{\sum_{i = 0}^{N} u_{i} s^{N-i}}}

        .. math::

            H_{z,n} = \\frac{\sum_{i = 0}^{N} b_{i} z^{-i}}{{1 - \sum_{i = 1}^{N} a_{i} z^{-i}}}

    **Example**

        ::

            from diamondback import ZTransform
            import math
            import numpy

            frequency, order, ripple = 0.1, 2, 0.125

            u = numpy.array( [ numpy.exp( 1j * math.pi * x / ( 2.0 * order ) ) for x in range( 1, 2 * order, 2 ) ] )

            v = math.asinh( 1.0 / ( ( 10.0 ** ( 0.1 * ripple ) - 1.0 ) ** 0.5 ) ) / order

            a = ( numpy.poly( ( -math.sinh( v ) * u.imag + 1j * math.cosh( v ) * u.real ) * 2.0 * math.pi ) ).real

            a /= a[ -1 ]

            # Transform z-domain coefficients with s-domain coefficients, frequency, and bilinear.

            a, b = ZTransform.transform( a = a, b = [ 1.0 ], frequency = frequency, bilinear = True )

            # Define zeros and normalize gain.

            b = numpy.poly( -numpy.ones( order ) )

            b *= ( 1.0 - sum( a ) ) / sum( b )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-26.

    **Definition**

"""

from typing import List, Tuple, Union
import math
import numpy
import scipy.signal

class ZTransform( object ) :

    """ Z transform realizes continuous s-domain to discrete z-domain transformation, through application of bilinear
        or impulse invariant methods.
    """

    @staticmethod
    def transform( u : Union[ List, numpy.ndarray ], v : Union[ List, numpy.ndarray ], frequency : float, bilinear : bool = True ) -> Tuple[ numpy.ndarray, numpy.ndarray ] :

        """ Transforms continuous s-domain coefficient arrays and produces
            discrete z-domain coefficient arrays.

            Arguments :

                u : Union[ List, numpy.ndarray ] - recursive coefficient, s-domain.

                v : Union[ List, numpy.ndarray ] - forward coefficient, s-domain.

                frequency : float - relative to Nyquist in ( 0.0, 1.0 ).

                bilinear : bool.

            Returns :

                a : numpy.ndarray - recursive coefficient, z-domain.

                b : numpy.ndarray - forward coefficient, z-domain.
        """

        if ( ( not numpy.isscalar( u ) ) and ( not isinstance( u, numpy.ndarray ) ) ) :

            u = numpy.array( list( u ) )

        if ( ( len( u.shape ) != 1 ) or ( len( u ) == 0 ) or ( not u.any( ) ) ) :

            raise ValueError( f'U = {u}' )

        if ( ( not numpy.isscalar( v ) ) and ( not isinstance( v, numpy.ndarray ) ) ) :

            v = numpy.array( list( v ) )

        if ( ( len( v.shape ) != 1 ) or ( len( v ) == 0 ) or ( not v.any( ) ) ) :

            raise ValueError( f'V = {v}' )

        if ( frequency <= 0.0 ) :

            raise ValueError( f'Frequency = {frequency}' )

        while ( not u[ 0 ] ) :

            u = numpy.delete( u, 0 )

        while ( not v[ 0 ] ) :

            v = numpy.delete( v, 0 )

        v /= u[ -1 ]

        u /= u[ -1 ]

        t = 1.0 / ( math.pi * frequency )

        if ( bilinear ) :

            p = numpy.roots( u )

            p = ( 1.0 + p / ( 2.0 * t ) ) / ( 1.0 - p / ( 2.0 * t ) )

            z = numpy.roots( v )

            z = ( 1.0 + z / ( 2.0 * t ) ) / ( 1.0 - z / ( 2.0 * t ) )

            if ( len( z ) < len( p ) ) :

                z = numpy.concatenate( ( z, numpy.zeros( len( p ) - len( z ) ) ) )

            a, b = numpy.poly( p ).real, numpy.poly( z ).real

        else :

            r, p, k = scipy.signal.residue( v, u )

            a, b = numpy.ones( 1 ) + 0j, 0j

            for ii in range( 0, len( r ) ) :

                a = numpy.convolve( a, numpy.array( [ 1.0, -numpy.exp( p[ ii ] / t ) ] ) )

                q = numpy.ones( 1 )

                for jj in range( 0, len( r ) ) :

                    if ( jj != ii ) :

                        q = numpy.convolve( q, numpy.array( [ 1.0, -numpy.exp( p[ jj ] / t ) ] ) )

                b += r[ ii ] * q

            if ( len( b ) < len( a ) ) :

                b = numpy.concatenate( ( b, numpy.zeros( len( a ) - len( b ) ) ) )

            a, b = a.real, b.real

        a /= -a[ 0 ]

        a[ 0 ] = 0.0

        b *= ( ( 1.0 - sum( a ) ) / sum( b ) )

        return a, b



""" **Description**

        A principal component model analyzes an incident signal to define
        transformation matrices which consume an incident signal to produce a
        reference signal, normalized and ordered to define orthogonal axes of
        descending variance.

        A principal component model is a supervised learning model which
        analyzes an incident signal representing a training set to learn a mean
        vector, standard deviation vector, and a collection of eigenvectors
        associated with an incident signal.

        .. math::

            \\vec{\mu_{i}} = \matrix{\ \\frac{\sum_{n=0}^{N}\\vec{x_{i,n}}}{N}}

        .. math::

            \\vec{\sigma_{i}} = \matrix{\ \\frac{\sum_{n=0}^{N}(\ \\vec{x_{i,n}} - \\vec{\mu_{i}}\ )^{2}}{N}}^{0.5}

        .. math::

            \Lambda_{n} = eig\matrix{\ cov\matrix{\ \matrix{\\frac{\ X_{n}^{T} - \\vec{\mu}\ }{\\vec{\sigma}}\ }\ }^{T}\ }^{T}

        An incident signal which is not a part of an inital training set is
        transformed without modifying a principal component model, by
        translation, normalization, and rotation to produce a reference signal
        which is a candidate for dimension reduction, in which higher order
        dimensions are discarded, reducing the order of the reference signal,
        while preserving significant and often sufficient information.

        .. math::

            Y_{n} = \Lambda_{n} \ \matrix{\\frac{\ X_{n}^{T} - \\vec{\mu}\ }{\\vec{\sigma}}\ }^{T}

        Principal component analysis and dimension reduction has application in
        clustering, classification, pattern recognition, and visualization.

    **Example**

        ::

            from diamondback import PrincipalComponentModel

            # Create an instance.

            obj = PrincipalComponentModel( )

            # Model an incident signal and extract eigenvalue, standard deviation, means, and rotation arrays.

            x = numpy.random.rand( 3, 32 )

            y = obj.model( x )

            e, s, u, v = obj.e, obj.s, obj.u, obj.v

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2019-01-25.

    **Definition**

"""

from diamondback.interfaces.IClear import IClear
from diamondback.interfaces.IEqual import IEqual
from diamondback.interfaces.IS import IS
from typing import Any, List, Union
import numpy

class PrincipalComponentModel( IClear, IS, IEqual ) :

    """ Principal component model.
    """

    @property
    def e( self ) :

        """ e : numpy.ndarray - eigenvalues.
        """

        return self._e

    @property
    def u( self ) :

        """ mean : numpy.ndarray.
        """

        return self._u

    @property
    def v( self ) :

        """ rotation : numpy.ndarray
        """

        return self._v

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.allclose( self.u, other.u ) ) and ( numpy.isclose( self.v, other.v ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._e, self._s, self._u, self._v = numpy.array( [ ] ), numpy.array( [ ] ), numpy.array( [ ] ), numpy.array( [ ] )

    def clear( self ) -> None :

        """ Clears an instance.
        """

        self._e, self._s, self._u, self._v = numpy.array( [ ] ), numpy.array( [ ] ), numpy.array( [ ] ), numpy.array( [ ] )

    def model( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Models an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal.

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 2 ) or ( len( x ) == 0 ) ) :

            raise ValueError( f'X = {x}' )

        if ( not len( self.s ) ) :

            self._s, self._u = x.std( 1 ), x.mean( 1 )

            z = ( ( x.T - self.u ) / self.s ).T

            self._e, self._v = numpy.linalg.eig( numpy.matmul( z, z.T ) / z.shape[ 1 ] )

            self._v = self._v.T

        else :

            z = ( ( x.T - self.u ) / self.s ).T

        rows, cols = x.shape

        if ( ( rows != len( self._u ) ) or ( ( rows, rows ) != self._v.shape ) or ( cols <= 0 ) ):

            raise ValueError( f'Rows = {rows} Columns = {cols}' )

        return numpy.matmul( self.v, z )



""" **Description**

        Time out interface.

    **Example**

        ::

            from diamondback import ITimeOut

            class Test( ITimeOut ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.timeout = ( 10.0, 60.0 )

            test = Test( )

            test.timeout = ( 15.0, 120.0 )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2019-10-09.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any

class ITimeOut( IEqual ) :

    """ Time out interface.
    """

    @property
    def timeout( self ) :

        """ timeout : Any.
        """

        return self._timeout

    @timeout.setter
    def timeout( self, timeout : Any ) :

        self._timeout = timeout

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( self.timeout == other.timeout ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._timeout = None



""" **Description**

        Phase interface.

    **Example**

        ::

            from diamondback import IPhase

            class Test( IPhase ) :

                def __init__( self ) -> None :

                    super( ).__init__( )

                    self.phase = 0.0

            test = Test( )

            test.phase = 0.5

    **License**

        `BSD-3C. <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.interfaces.IEqual import IEqual
from typing import Any
import numpy

class IPhase( IEqual ) :

    """ Phase interface.
    """

    @property
    def phase( self ) :

        """ phase : float - relative to pi in [ -1.0, 1.0 ].
        """

        return self._phase

    @phase.setter
    def phase( self, phase : float ) :

        if ( ( phase < -1.0 ) or ( phase > 1.0 ) ) :

            raise ValueError( f'Phase = {phase}' )

        self._phase = phase

    def __eq__( self, other : Any ) -> bool :

        """ Equal.

            Arguments :

                other : Any.

            Returns :

                equal : bool.
        """

        return ( ( super( ).__eq__( other ) ) and ( numpy.isclose( self.phase, other.phase ) ) )

    def __init__( self ) -> None :

        """ Initialize.
        """

        super( ).__init__( )

        self._phase = 0.0



""" **Description**

        A complex exponential filter produces a complex exponential reference
        signal from an incident signal equal to a specified normalized frequency.
        A normalized phase is specified.

        .. math::

            x_{n} = f_{n}

        .. math::

            \phi_{n} = \phi_{n-1} + x_{n}

        .. math::

            y_{n} = e^{\ j\ \pi\ \phi_{n}}

    **Example**

        ::

            from diamondback import ComplexExponentialFilter
            import numpy

            # Create an instance with phase.

            obj = ComplexExponentialFilter( phase = 0.0 )

            # Filter an incident signal.

            x = numpy.linspace( -1.0e-4, 1.0e-4, 128 ) + 0.1

            y = obj.filter( x )

    **License**

        `BSD-3C.  <https://github.com/larryturner/diamondback/blob/master/license>`_

        © 2018 - 2021 Larry Turner, Schneider Electric Industries SAS. All rights reserved.

    **Author**

        Larry Turner, Schneider Electric, Analytics & AI, 2018-01-31.

    **Definition**

"""

from diamondback.interfaces.IPhase import IPhase
from typing import List, Union
import math
import numpy

class ComplexExponentialFilter( IPhase ) :

    """ Complex exponential filter.
    """

    def __init__( self, phase : float = 0.0 ) -> None :

        """ Initialize.

            Arguments :

                phase : float - relative to pi in [ -1.0, 1.0 ].
        """

        super( ).__init__( )

        self.phase = phase

    def filter( self, x : Union[ List, numpy.ndarray ] ) -> numpy.ndarray :

        """ Filters an incident signal and produces a reference signal.

            Arguments :

                x : Union[ List, numpy.ndarray ] - incident signal frequency relative to Nyquist in [ -1.0, 1.0 ).

            Returns :

                y : numpy.ndarray - reference signal.
        """

        if ( ( not numpy.isscalar( x ) ) and ( not isinstance( x, numpy.ndarray ) ) ) :

            x = numpy.array( list( x ) )

        if ( ( len( x.shape ) != 1 ) or ( len( x ) == 0 ) or ( abs( x ) > 1.0 ).any( ) ) :

            raise ValueError( f'X = {x}' )

        y = numpy.zeros( len( x ), complex )

        phase = self.phase

        for ii in range( 0, len( x ) ) :

            y[ ii ] = numpy.exp( 1j * math.pi * phase )

            phase += x[ ii ]

        phase = numpy.fmod( phase, 2.0 )

        if ( abs( phase ) > 1.0 ) :

            phase -= numpy.sign( phase ) * 2.0

        self.phase = phase

        return y
